/* #includes */ /*{{{C}}}*//*{{{*/
#undef  _POSIX_SOURCE
#define _POSIX_SOURCE   1
#undef  _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 2

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>

#ifdef USE_DMALLOC
#include "dmalloc.h"
#endif

#include "getopt.h"
#include "config.h"

#include "bas.h"
/*}}}*/

int main(int argc, char *argv[]) /*{{{*/
{
  /* variables */ /*{{{*/
  char *runFile=(char*)0;
  int usage=0,o;
  const char *lp="/dev/null";
  int backslash_colon=0;
  int lpfd;
  static struct option lopts[]=
  {
    { "lp", required_argument, 0, 'l' },
    { "help", no_argument, 0, 'h' },
    { "vax-basic", no_argument, 0, 'v' },
    { "version", no_argument, 0, 'V' },
#if defined(__STDC__) && __STDC__
    { (const char*)0, 0, 0, '\0' }
#else
    { (char*)0, 0, 0, '\0' }
#endif
  };
  /*}}}*/

  /* parse arguments */ /*{{{*/
  while ((o=getopt_long(argc,argv,"bl:h?",lopts,(int*)0))!=EOF) switch (o)
  {
    case 'b': backslash_colon=1; break;
    case 'l': lp=optarg; break;
    case 'V': printf("bas " VERSION "\n"); exit(0); break;
    case 'h': usage=2; break;
    default: usage=1; break;
  }
  if (optind<argc) runFile=argv[optind++];
  if (usage==0 && optind!=argc) usage=1;
  if (usage==1)
  {
    fputs("Usage: bas [-b] [-l file] [program]\n",stderr);
    fputs("       bas [--backslash-colon] [--lp file] [program]\n",stderr);
    fputs("       bas -h|--help\n",stderr);
    fputs("       bas --version\n",stderr);
    fputs("\n",stderr);
    fputs("Try `bas -h' or `bas --help' for more information.\n",stderr);
    exit(1);
  }
  if (usage==2)
  {
    fputs("Usage: bas [-b] [-l file] [program]\n",stdout);
    fputs("       bas [--backslash-colon] [--lp file] [program]\n",stdout);
    fputs("       bas -h|--help\n",stdout);
    fputs("       bas --version\n",stdout);
    fputs("\n",stdout);
    fputs("BASIC interpreter.\n",stdout);
    fputs("\n",stdout);
    fputs("-b, --backslash-colon  convert backslashs to colons\n",stdout);
    fputs("-l, --lp               write LPRINT output to file\n",stdout);
    fputs("-h, --help             display this help and exit\n",stdout);
    fputs("    --version          output version information and exit\n",stdout);
    fputs("\n",stdout);
    fputs("Report bugs to <michael@moria.de>.\n",stdout);
    exit(0);
  }
  if ((lpfd=open(lp,O_WRONLY|O_CREAT|O_TRUNC,0666))==-1)
  {
    fprintf(stderr,"bas: unable to open %s for line printer output (%s).\n",lp,strerror(errno));
    exit(2);
  }
  /*}}}*/
  bas_init(backslash_colon,lpfd);
  if (runFile) bas_runFile(runFile);
  else bas_interpreter();
  bas_exit();
  return(0);
}
/*}}}*/
