/* Tokens and token sequence arrays. */
%{
/* #includes */ /*{{{C}}}*//*{{{*/
#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <math.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include "auto.h"
#include "token.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif
/*}}}*/

static int matchdata;
static int backslash_colon;
int yylex(void);
static struct Token *cur;

static void string(void) /*{{{*/
{
  char *t,*q;
  size_t l;

  if (cur)
  {
    for (t=yytext+1,l=0; *(t+1); ++t,++l)
    {
      if (*t=='"') ++t;
    }
    cur->u.string=malloc(sizeof(struct String));
    String_size(String_new(cur->u.string),l);
    for (t=yytext+1,q=cur->u.string->character; *(t+1); ++t,++q)
    {
      *q=*t;
      if (*t=='"') ++t;
    }
  }
}
/*}}}*/
static void string2(void) /*{{{*/
{
  char *t,*q;
  size_t l;

  if (cur)
  {
    for (t=yytext+1,l=0; *t; ++t,++l)
    {
      if (*t=='"') ++t;
    }
    cur->u.string=malloc(sizeof(struct String));
    String_size(String_new(cur->u.string),l);
    for (t=yytext+1,q=cur->u.string->character; *t; ++t,++q)
    {
      *q=*t;
      if (*t=='"') ++t;
    }
  }
}
/*}}}*/
static void setBit(int *field, int bit) /*{{{*/
{
  field[bit/(sizeof(int)*8)]|=1<<(bit&((sizeof(int)*8)-1));
}
/*}}}*/
%}
 /* flex options and definitions */ /*{{{*/
%option noyywrap
%option nounput
%x DATAINPUT ELSEIF
REAL		[0-9]*(((\.[0-9]+(e("+"|"-")?[0-9]+)?)?("!"|"#"))|(\.[0-9]+(e("+"|"-")?[0-9]+)?("!"|"#")?))
INTEGER		[0-9]+%?
HEXINTEGER	&H[0-9A-F]+
IDENTIFIER      [A-Z][A-Z_0-9\.]*("$"|"%"|"#")?
STRING          \"([^"]|\"\")*\"
STRING2		\"([^"]|\"\")*$
REM		rem([^0-9A-Z_\.\n][^\n]*)?
QUOTE		("'"|"!")[^\n]*
ENDIF           end[ \t]*if
ENDPROC         end[ \t]*proc
ENDSUB          end[ \t]+sub
ENDSELECT	end[ \t]*select
DOWHILE		do[ \t]+while
EXITDO		exit[ \t]+do
EXITFOR		exit[ \t]+for
GOTO		go[ \t]*to
LINEINPUT	(line[ \t]+input)|linput
LOOPUNTIL	loop[ \t]+until
DATAITEM	[^ \t\n,:][^,:\n]*
ONERROR		on[ \t]+error
ONERROROFF	on[ \t]+error[ \t]+off
ONERRORGOTO0	on[ \t]+error[ \t]+goto[ \t]+0
SELECTCASE	select[ \t]+case
 /*}}}*/
%%
 /* flex rules */ /*{{{*/
  if (matchdata) BEGIN(DATAINPUT);

"#"			return T_CHANNEL;
{REAL}          	{
                	  if (cur) cur->u.real=strtod(yytext,(char**)0);
                	  return T_REAL;
                	}
{INTEGER}       	{
                	  if (cur) cur->u.integer=strtol(yytext,(char**)0,10);
                  	  return T_INTEGER;
                	}
{HEXINTEGER}    	{
                	  if (cur) cur->u.hexinteger=strtoul(yytext+2,(char**)0,16);
                	  return T_HEXINTEGER;
                	}
{STRING}        	string(); return T_STRING;
{STRING2}       	string2(); return T_STRING;
"("             	return T_OP;
")"             	return T_CP;
"*"             	return T_MULT;
"+"             	return T_PLUS;
"-"             	return T_MINUS;
","             	return T_COMMA;
"/"             	return T_DIV;
"\\"			return backslash_colon?T_COLON:T_IDIV;
":"             	return T_COLON;
";"             	return T_SEMICOLON;
"<"             	return T_LT;
"<="            	return T_LE;
"=<"            	return T_LE;
"<>"            	return T_NE;
"="             	return T_EQ;
">"             	return T_GT;
">="            	return T_GE;
"=>"            	return T_GE;
"^"             	return T_POW;
"and"			return T_AND;
"as"			return T_AS;
"call"			return T_CALL;
"case"[ \t]+"else"	{
			  if (cur) cur->u.casevalue=malloc(sizeof(struct Casevalue));
			  return T_CASEELSE;
			}
"case"			{
			  if (cur) cur->u.casevalue=malloc(sizeof(struct Casevalue));
			  return T_CASEVALUE;
			}
"chdir"			return T_CHDIR;
"clear"         	return T_CLEAR;
"close"			return T_CLOSE;
"close"/"#"		return T_CLOSE;
"cls"			return T_CLS;
"con"			return T_CON;
"data"|"d."		BEGIN(DATAINPUT); return T_DATA;
<DATAINPUT>{STRING}	string(); return T_STRING;
<DATAINPUT>{STRING2}	string2(); return T_STRING;
<DATAINPUT>","		return T_COMMA;
<DATAINPUT>{DATAITEM}	{
	                  if (cur) cur->u.datainput=strcpy(malloc(strlen(yytext)+1),yytext);
			  return T_DATAINPUT;
			}
<DATAINPUT>[ \t]+
<DATAINPUT>\n		BEGIN(INITIAL);
<DATAINPUT>:		BEGIN(INITIAL); return T_COLON;
"defdbl"        	return T_DEFDBL;
"defint"        	return T_DEFINT;
"defstr"		return T_DEFSTR;
"def"/[ \t]+fn[A-Z_0-9\.] return T_DEFFN;
"def"/[ \t]+proc[A-Z_0-9\.] return T_DEFPROC;
"delete"		return T_DELETE;
"dim"           	return T_DIM;
"do"			return T_DO;
{DOWHILE}		return T_DOWHILE;
"edit"			return T_EDIT;
"else"|"el."		return T_ELSE;
"else"/"if"		BEGIN(ELSEIF); return T_ELSEIFELSE;
<ELSEIF>"if"		BEGIN(INITIAL); return T_ELSEIFIF;
end[ \t]+function	return T_ENDFN;
{ENDIF}         	return T_ENDIF;
{ENDPROC}		return T_ENDPROC;
{ENDSUB}		return T_ENDSUB;
{ENDSELECT}		return T_ENDSELECT;
"end"           	return T_END;
"environ"		return T_ENVIRON;
"erase"         	return T_ERASE;
"eqv"			return T_EQV;
{EXITDO}		return T_EXITDO;
{EXITFOR}		return T_EXITFOR;
"field"			return T_FIELD;
"field"/"#"		return T_FIELD;
"file"			return T_FILE;
"fnend"			return T_FNEND;
"fnreturn"		return T_FNRETURN;
"for"           	return T_FOR;
"function"		return T_FUNCTION;
"get"			return T_GET;
"get"/"#"		return T_GET;
"gosub"         	return T_GOSUB;
{GOTO}          	return T_GOTO;
"idn"			return T_IDN;
"if"            	return T_IF;
"imp"			return T_IMP;
"input"			return T_INPUT;
"input"/"#"		return T_INPUT;
"inv"			return T_INV;
"kill"			return T_KILL;
"let"           	return T_LET;
"list"          	return T_LIST;
"llist"			return T_LLIST;
"load"          	return T_LOAD;
"local"			return T_LOCAL;
"locate"		return T_LOCATE;
"loop"			return T_LOOP;
{LOOPUNTIL}		return T_LOOPUNTIL;
"lprint"		return T_LPRINT;
"lset"			return T_LSET;
"mat"[ \t]+"input"	return T_MATINPUT;
"mat"[ \t]+"print"	return T_MATPRINT;
"mat"[ \t]+"read"	return T_MATREAD;
"mat"			return T_MAT;
"mkdir"			return T_MKDIR;
"mod"           	return T_MOD;
"new"           	return T_NEW;
"name"			return T_NAME;
"next"          	{
                  	  if (cur) cur->u.next=malloc(sizeof(struct Next));
			  return T_NEXT;
                	}
"not"			return T_NOT;
{ONERROROFF}		return T_ONERROROFF;
{ONERRORGOTO0}		return T_ONERRORGOTO0;
{ONERROR}		return T_ONERROR;
"on"			{
			  if (cur)
			  {
			    cur->u.on.pcLength=1;
			    cur->u.on.pc=(struct Pc*)0;
			  }
			  return T_ON;
			}
"open"			return T_OPEN;
"or"			return T_OR;
"output"		return T_OUTPUT;
"print"|"p."|"?"	return T_PRINT;
("print"|"p."|"?")/"#"	return T_PRINT;
"put"			return T_PUT;
"put"/"#"		return T_PUT;
"random"		return T_RANDOM;
"randomize"     	return T_RANDOMIZE;
"read"			return T_READ;
"renum"|"ren."		return T_RENUM;
"repeat"|"rep."		return T_REPEAT;
"restore"|"res."	return T_RESTORE;
"resume"		return T_RESUME;
"return"|"r."   	return T_RETURN;
"rset"			return T_RSET;
"run"           	return T_RUN;
"save"			return T_SAVE;
{SELECTCASE}		{
			  if (cur) cur->u.selectcase=malloc(sizeof(struct Selectcase));
			  return T_SELECTCASE;
			}
"shell"			return T_SHELL;
"spc"			return T_SPC;
"step"          	return T_STEP;
"stop"          	return T_STOP;
"sub"			return T_SUB;
"swap"			return T_SWAP;
"system"        	return T_SYSTEM;
"then"|"th."		return T_THEN;
"tab"			return T_TAB;
"to"            	return T_TO;
"trn"			return T_TRN;
"troff"         	return T_TROFF;
"tron"          	return T_TRON;
"unnum"			return T_UNNUM;
"until"         	return T_UNTIL;
"using"			return T_USING;
"wend"          	{
                	  if (cur) cur->u.whilepc=malloc(sizeof(struct Pc));
                	  return T_WEND;
                	}
"while"         	{
                	  if (cur) cur->u.afterwend=malloc(sizeof(struct Pc));
                	  return T_WHILE;
                	}
"width"			return T_WIDTH;
"width"/"#"		return T_WIDTH;
"write"			return T_WRITE;
"write"/"#"		return T_WRITE;
"xor"			return T_XOR;
"zer"			return T_ZER;
{REM}           	{
                	  if (cur) cur->u.rem=strcpy(malloc(strlen(yytext+3)+1),yytext+3);
                	  return T_REM;
                	}
{QUOTE}          	{
                	  if (cur) strcpy(cur->u.rem=malloc(strlen(yytext+1)+1),yytext+1);
                	  return T_QUOTE;
                	}
{LINEINPUT}		return T_LINEINPUT;
{IDENTIFIER}		{
			  if (cur)
			  {
			    cur->u.identifier=malloc(offsetof(struct Identifier,name)+strlen(yytext)+1);
			    strcpy(cur->u.identifier->name,yytext);
			    switch (yytext[yyleng-1])
			    {
			      case '$': cur->u.identifier->defaultType=V_STRING; break;
			      case '%': cur->u.identifier->defaultType=V_INTEGER; break;
			      default: cur->u.identifier->defaultType=V_REAL; break;
			    }
			  }
			  return T_IDENTIFIER;
			}
[ \t\n]+
.			{
			  if (cur) cur->u.junk=yytext[0];
			  return T_JUNK;
			}
 /*}}}*/
%%

int Token_property[T_LASTTOKEN];

struct Token *Token_newCode(const char *ln) /*{{{*/
{
  int l,lasttok,thistok,addNumber=0,sawif;
  struct Token *result;
  YY_BUFFER_STATE buf;

  cur=(struct Token*)0;
  buf=yy_scan_string(ln);
  matchdata=sawif=0;
  for (lasttok=T_EOL,l=1; (thistok=yylex()); ++l)
  {
    if (l==1 && thistok!=T_INTEGER) { addNumber=1; ++l; }
    if ((lasttok==T_THEN || lasttok==T_ELSE) && thistok==T_INTEGER) ++l;
    if (thistok==T_IF) sawif=1;
    if (thistok==T_THEN) sawif=0;
    if (thistok==T_GOTO && sawif) ++l;
    lasttok=thistok;
  }
  if (l==1) { addNumber=1; ++l; }
  yy_delete_buffer(buf);
  cur=result=malloc(sizeof(struct Token)*l);
  if (addNumber)
  {
    cur->type=T_UNNUMBERED;
    ++cur;
  }
  buf=yy_scan_string(ln);
  lasttok=T_EOL;
  matchdata=sawif=0;
  while ((cur->type=yylex()))
  {
    if (cur->type==T_IF) sawif=1;
    if (cur->type==T_THEN) sawif=0;
    if (cur->type==T_GOTO && sawif)
    {
      sawif=0;
      *(cur+1)=*cur;
      cur->type=T_THEN;
      lasttok=T_GOTO;
      cur+=2;
    }
    else if ((lasttok==T_THEN || lasttok==T_ELSE) && cur->type==T_INTEGER)
    {
      *(cur+1)=*cur;
      cur->type=T_GOTO;
      lasttok=T_INTEGER;
      cur+=2;
    }
    else
    {
      lasttok=cur->type;
      ++cur;
    }
  }
  cur->type=T_EOL;
  yy_delete_buffer(buf);
  return result;
}
/*}}}*/
struct Token *Token_newData(const char *ln) /*{{{*/
{
  int l;
  struct Token *result;
  YY_BUFFER_STATE buf;

  cur=(struct Token*)0;
  buf=yy_scan_string(ln);
  matchdata=1;
  for (l=1; yylex(); ++l);
  yy_delete_buffer(buf);
  cur=result=malloc(sizeof(struct Token)*l);
  buf=yy_scan_string(ln);
  matchdata=1;
  while ((cur->type=yylex())) ++cur;
  cur->type=T_EOL;
  yy_delete_buffer(buf);
  return result;
}
/*}}}*/
void Token_destroy(struct Token *token) /*{{{*/
{
  struct Token *r=token;

  do
  {
    switch (r->type)
    {
      case T_AND:          break;
      case T_AS:           break;
      case T_CALL:         break;
      case T_CASEELSE:
      case T_CASEVALUE:    free(r->u.casevalue); break;
      case T_CHANNEL:      break;
      case T_CHDIR:        break;
      case T_CLEAR:        break;
      case T_CLOSE:        break;
      case T_CLS:          break;
      case T_COLON:        break;
      case T_COMMA:        break;
      case T_CON:          break;
      case T_CP:           break;
      case T_DATA:         break;
      case T_DATAINPUT:    free(r->u.datainput); break;
      case T_DEFFN:        break;
      case T_DEFDBL:       break;
      case T_DEFINT:       break;
      case T_DEFPROC:      break;
      case T_DEFSTR:       break;
      case T_DELETE:       break;
      case T_DIM:          break;
      case T_DIV:          break;
      case T_DO:           break;
      case T_DOWHILE:      break;
      case T_EDIT:         break;
      case T_ELSE:         break;
      case T_ELSEIFELSE:   break;
      case T_ELSEIFIF:     break;
      case T_END:          break;
      case T_ENDFN:        break;
      case T_ENDIF:        break;
      case T_ENDPROC:      break;
      case T_ENDSELECT:    break;
      case T_ENDSUB:       break;
      case T_ENVIRON:      break;
      case T_EOL:          break;
      case T_EQ:           break;
      case T_EQV:          break;
      case T_ERASE:        break;
      case T_EXITDO:       break;
      case T_EXITFOR:      break;
      case T_FIELD:        break;
      case T_FILE:         break;
      case T_FNEND:        break;
      case T_FNRETURN:     break;
      case T_FOR:          break;
      case T_FUNCTION:     break;
      case T_GE:           break;
      case T_GET:          break;
      case T_GOSUB:        break;
      case T_GOTO:         break;
      case T_GT:           break;
      case T_HEXINTEGER:   break;
      case T_IDENTIFIER:   free(r->u.identifier); break;
      case T_IDIV:         break;
      case T_IDN:          break;
      case T_IF:           break;
      case T_IMP:          break;
      case T_INPUT:        break;
      case T_INTEGER:      break;
      case T_INV:          break;
      case T_JUNK:         break;
      case T_KILL:         break;
      case T_LE:           break;
      case T_LET:          break;
      case T_LINEINPUT:    break;
      case T_LIST:         break;
      case T_LLIST:        break;
      case T_LOAD:         break;
      case T_LOCAL:        break;
      case T_LOCATE:       break;
      case T_LOOP:         break;
      case T_LOOPUNTIL:    break;
      case T_LPRINT:       break;
      case T_LSET:         break;
      case T_LT:           break;
      case T_MAT:          break;
      case T_MATINPUT:     break;
      case T_MATPRINT:     break;
      case T_MATREAD:      break;
      case T_MINUS:        break;
      case T_MKDIR:        break;
      case T_MOD:          break;
      case T_MULT:         break;
      case T_NAME:         break;
      case T_NE:           break;
      case T_NEW:          break;
      case T_NEXT:         free(r->u.next); break;
      case T_NOT:          break;
      case T_ON:           if (r->u.on.pc) free(r->u.on.pc); break;
      case T_ONERROR:      break;
      case T_ONERRORGOTO0: break;
      case T_ONERROROFF:   break;
      case T_OP:           break;
      case T_OPEN:         break;
      case T_OR:           break;
      case T_OUTPUT:       break;
      case T_PLUS:         break;
      case T_POW:          break;
      case T_PRINT:        break;
      case T_PUT:          break;
      case T_QUOTE:        free(r->u.rem); break;
      case T_RANDOM:       break;
      case T_RANDOMIZE:    break;
      case T_READ:         break;
      case T_REAL:         break;
      case T_REM:          free(r->u.rem); break;
      case T_RENUM:        break;
      case T_REPEAT:       break;
      case T_RESTORE:      break;
      case T_RESUME:       break;
      case T_RETURN:       break;
      case T_RSET:         break;
      case T_RUN:          break;
      case T_SAVE:         break;
      case T_SELECTCASE:   free(r->u.selectcase); break;
      case T_SEMICOLON:    break;
      case T_SHELL:        break;
      case T_SPC:          break;
      case T_STEP:         break;
      case T_STOP:         break;
      case T_STRING:       String_destroy(r->u.string); free(r->u.string); break;
      case T_SUB:          break;
      case T_SWAP:         break;
      case T_SYSTEM:       break;
      case T_TAB:          break;
      case T_THEN:         break;
      case T_TO:           break;
      case T_TRN:          break;
      case T_TROFF:        break;
      case T_TRON:         break;
      case T_UNNUM:        break;
      case T_UNNUMBERED:   break;
      case T_UNTIL:        break;
      case T_USING:        break;
      case T_WEND:         free(r->u.whilepc); break;
      case T_WHILE:        free(r->u.afterwend); break;
      case T_WIDTH:        break;
      case T_WRITE:        break;
      case T_XOR:          break;
      case T_ZER:          break;
      default:             assert(0);
    }
  } while ((r++)->type!=T_EOL);
  free(token);
}
/*}}}*/
struct String *Token_toString(struct Token *token, struct Token *spaceto, struct String *s, int *indent, int width) /*{{{*/
{
  int ns=0,infn=0;
  int thisindent=0,nextindent=0;
  size_t oldlength=s->length;
  struct Token *t;
  static struct
  {
    const char *text;
    char space;
  } table[]=
  {
    /* 0              */ {(const char*)0,-1},
    /* T_AND          */ {"and",1},
    /* T_AS           */ {"as",1},
    /* T_CALL         */ {"call",1},
    /* T_CASEELSE     */ {"case else",1},
    /* T_CASEVALUE    */ {"case",1},
    /* T_CHANNEL      */ {"#",0},
    /* T_CHDIR        */ {"chdir",1},
    /* T_CLEAR        */ {"clear",1},
    /* T_CLOSE        */ {"close",1},
    /* T_CLS          */ {"cls",1},
    /* T_COLON        */ {":",1},
    /* T_COMMA        */ {",",0},
    /* T_CON          */ {"con",0},
    /* T_CP           */ {")",0},
    /* T_DATA         */ {"data",1},
    /* T_DATAINPUT    */ {(const char*)0,0},
    /* T_DEFDBL       */ {"defdbl",1},
    /* T_DEFFN        */ {"def",1},
    /* T_DEFINT       */ {"defint",1},
    /* T_DEFPROC      */ {"def",1},
    /* T_DEFSTR       */ {"defstr",1},
    /* T_DELETE       */ {"delete",1},
    /* T_DIM          */ {"dim",1},
    /* T_DIV          */ {"/",0},
    /* T_DO           */ {"do",1},
    /* T_DOWHILE      */ {"do while",1},
    /* T_EDIT         */ {"edit",1},
    /* T_ELSE         */ {"else",1},
    /* T_ELSEIFELSE   */ {"elseif",1},
    /* T_ELSEIFIF     */ {(const char*)0,0},
    /* T_END          */ {"end",1},
    /* T_ENDFN        */ {"end function",1},
    /* T_ENDIF        */ {"end if",1},
    /* T_ENDPROC      */ {"end proc",1},
    /* T_ENDSELECT    */ {"end select",1},
    /* T_ENDSUB       */ {"end sub",1},
    /* T_ENVIRON      */ {"environ",1},
    /* T_EOL          */ {"\n",0},
    /* T_EQ           */ {"=",0},
    /* T_EQV          */ {"eqv",0},
    /* T_ERASE        */ {"erase",1},
    /* T_EXITDO       */ {"exit do",1},
    /* T_EXITFOR      */ {"exit for",1},
    /* T_FIELD        */ {"field",1},
    /* T_FILE         */ {"file",1},
    /* T_FNEND        */ {"fnend",1},
    /* T_FNRETURN     */ {"fnreturn",1},
    /* T_FOR          */ {"for",1},
    /* T_FUNCTION     */ {"function",1},
    /* T_GE           */ {">=",0},
    /* T_GET          */ {"get",1},
    /* T_GOSUB        */ {"gosub",1},
    /* T_GOTO         */ {"goto",1},
    /* T_GT           */ {">",0},
    /* T_HEXINTEGER   */ {(const char*)0,0},
    /* T_IDENTIFIER   */ {(const char*)0,0},
    /* T_IDIV         */ {"\\",0},
    /* T_IDN          */ {"idn",0},
    /* T_IF           */ {"if",1},
    /* T_IMP          */ {"imp",0},
    /* T_INPUT        */ {"input",1},
    /* T_INTEGER      */ {(const char*)0,0},
    /* T_INV          */ {"inv",0},
    /* T_JUNK         */ {(const char*)0,1},
    /* T_KILL         */ {"kill",1},
    /* T_LE           */ {"<=",0},
    /* T_LET          */ {"let",1},
    /* T_LINEINPUT    */ {"line input",1},
    /* T_LIST         */ {"list",1},
    /* T_LLIST        */ {"llist",1},
    /* T_LOAD         */ {"load",1},
    /* T_LOCAL        */ {"local",1},
    /* T_LOCATE       */ {"locate",1},
    /* T_LOOP         */ {"loop",1},
    /* T_LOOPUNTIL    */ {"loop until",1},
    /* T_LPRINT       */ {"lprint",1},
    /* T_LSET         */ {"lset",1},
    /* T_LT           */ {"<",0},
    /* T_MAT          */ {"mat",1},
    /* T_MATINPUT     */ {"mat input",1},
    /* T_MATPRINT     */ {"mat print",1},
    /* T_MATREAD      */ {"mat read",1},
    /* T_MINUS        */ {"-",0},
    /* T_MKDIR        */ {"mkdir",1},
    /* T_MOD          */ {"mod",0},
    /* T_MULT         */ {"*",0},
    /* T_NAME         */ {"name",1},
    /* T_NE           */ {"<>",0},
    /* T_NEW          */ {"new",1},
    /* T_NEXT         */ {"next",1},
    /* T_NOT          */ {"not",0},
    /* T_ON           */ {"on",1},
    /* T_ONERROR      */ {"on error",1},
    /* T_ONERRORGOTO0 */ {"on error goto 0",1},
    /* T_ONERROROFF   */ {"on error off",1},
    /* T_OP           */ {"(",0},
    /* T_OPEN         */ {"open",1},
    /* T_OR           */ {"or",1},
    /* T_OUTPUT       */ {"output",1},
    /* T_PLUS         */ {"+",0},
    /* T_POW          */ {"^",0},
    /* T_PRINT        */ {"print",1},
    /* T_PUT          */ {"put",1},
    /* T_QUOTE        */ {(const char*)0,1},
    /* T_RANDOM       */ {"random",1},
    /* T_RANDOMIZE    */ {"randomize",1},
    /* T_READ         */ {"read",1},
    /* T_REAL         */ {(const char*)0,0},
    /* T_REM          */ {(const char*)0,1},
    /* T_RENUM        */ {"renum",1},
    /* T_REPEAT       */ {"repeat",1},
    /* T_RESTORE      */ {"restore",1},
    /* T_RESUME       */ {"resume",1},
    /* T_RETURN       */ {"return",1},
    /* T_RSET         */ {"rset",1},
    /* T_RUN          */ {"run",1},
    /* T_SAVE         */ {"save",1},
    /* T_SELECTCASE   */ {"select case",1},
    /* T_SEMICOLON    */ {";",0},
    /* T_SHELL        */ {"shell",1},
    /* T_SPC          */ {"spc",0},
    /* T_STEP         */ {"step",1},
    /* T_STOP         */ {"stop",1},
    /* T_STRING       */ {(const char*)0,0},
    /* T_SUB          */ {"sub",1},
    /* T_SWAP         */ {"swap",1},
    /* T_SYSTEM       */ {"system",1},
    /* T_TAB          */ {"tab",0},
    /* T_THEN         */ {"then",1},
    /* T_TO           */ {"to",1},
    /* T_TRN          */ {"trn",0},
    /* T_TROFF        */ {"troff",1},
    /* T_TRON         */ {"tron",1},
    /* T_UNNUM        */ {"unnum",1},
    /* T_UNNUMBERED   */ {"",0},
    /* T_UNTIL        */ {"until",1},
    /* T_USING        */ {"using",0},
    /* T_WEND         */ {"wend",1},
    /* T_WHILE        */ {"while",1},
    /* T_WIDTH        */ {"width",1},
    /* T_WRITE        */ {"write",1},
    /* T_XOR          */ {"xor",0},
    /* T_ZER          */ {"zer",0},
  };

  /* precompute indentation */ /*{{{*/
  if (indent) { thisindent=*indent; nextindent=*indent; }
  t=token;
  do
  {
    switch (t->type)
    {
      case T_CASEELSE:
      case T_CASEVALUE:
      {
        if (thisindent) --thisindent;
        break;
      }
      case T_DEFFN:
      case T_FUNCTION:
      {
        struct Token *cp;

        for (cp=t; cp->type!=T_EOL && cp->type!=T_CP; ++cp);
        if ((cp+1)->type!=T_EQ) ++nextindent;
        infn=1;
        break;
      }
      case T_COLON: infn=0; break;
      case T_DEFPROC:
      case T_DO:
      case T_DOWHILE:
      case T_FOR:
      case T_REPEAT:
      case T_SUB:
      case T_WHILE: ++nextindent; break;
      case T_SELECTCASE: nextindent+=2; break;
      case T_EQ:
      {
        if (infn || (t>token && ((t-1)->type==T_COLON || (t-1)->type==T_INTEGER || (t-1)->type==T_UNNUMBERED)))
        {
          if (thisindent) --thisindent;
          if (nextindent) --nextindent;
        }
        infn=0;
        break;
      }
      case T_ENDFN:
      case T_FNEND:
      case T_ENDIF:
      case T_ENDPROC:
      case T_ENDSUB:
      case T_LOOP:
      case T_UNTIL:
      case T_WEND:
      {
        if (thisindent) --thisindent;
        if (nextindent) --nextindent;
        break;
      }
      case T_ENDSELECT:
      {
        if (thisindent) --thisindent;
        if (thisindent) --thisindent;
        if (nextindent) --nextindent;
        if (nextindent) --nextindent;
        break;
      }
      case T_NEXT:
      {
        ++t;
        while (1)
        {
          if (thisindent) --thisindent;
          if (nextindent) --nextindent;
          if (t->type==T_IDENTIFIER)
          {
            int par=0;

            ++t;
            if (t->type==T_OP)
            {
              do
              {
                if (t->type==T_OP) ++par;
                else if (t->type==T_CP) --par;
                if (t->type!=T_EOL) ++t;
                else break;
              } while (par);
            }
            if (t->type==T_COMMA) ++t;
            else break;
          }
          else break;
        }
        break;
      }
      case T_THEN: if ((t+1)->type==T_EOL) ++nextindent; break;
      case T_ELSE: if (t==token+1 && thisindent) --thisindent; break;
      case T_ELSEIFELSE:
      {
        if (t==token+1 && thisindent) --thisindent;
        if (nextindent) --nextindent;
        break;
      }
      default: break;
    }
  } while (t++->type!=T_EOL);
  /*}}}*/
  if (width>=0) /* whole line */
  {
    if (width) /* nicely formatted listing */
    {
      assert (token->type==T_UNNUMBERED || token->type==T_INTEGER);
      if (token->type==T_INTEGER) String_appendPrintf(s,"%*ld ",width,token->u.integer);
      else String_appendPrintf(s,"%*s ",width,"");
    }
    else assert (token->type==T_UNNUMBERED);
    ++token;
  }
  while (thisindent--) String_appendPrintf(s,"  ");
  do
  {

    if (s->length && token->type!=T_EOL)
    {
      const char *keyword;

      if ((keyword=table[token->type].text)==(const char*)0) keyword="X";
      if (ns && s->character[s->length-1]!=' ')
      {
        String_appendPrintf(s," ");
      }
      else if (isalnum(s->character[s->length-1]) && isalnum(*keyword))
      {
        String_appendPrintf(s," ");
      }
      else if (s->character[s->length-1]!=' ' && table[token->type].space)
      {
        String_appendChar(s,' ');
      }
    }
    if (spaceto && token==spaceto) break;
    switch (token->type)
    {
      case T_DATAINPUT: String_appendChars(s,token->u.datainput); break;
      case T_ELSEIFIF: break;
      case T_IDENTIFIER: String_appendChars(s,token->u.identifier->name); break;
      case T_INTEGER: String_appendPrintf(s,"%ld",token->u.integer); break;
      case T_HEXINTEGER: String_appendPrintf(s,"&h%lx",token->u.hexinteger); break;
      case T_JUNK: String_appendChar(s,token->u.junk); break;
      case T_REAL:
      {
        String_appendPrintf(s,"%.*g",DBL_DIG,token->u.real);
        if ((token->u.real<((double)LONG_MIN)) || (token->u.real>((double)LONG_MAX))) String_appendChar(s,'!');
        break;
      }
      case T_REM: String_appendPrintf(s,"rem%s",token->u.rem); break;
      case T_QUOTE: String_appendPrintf(s,"'%s",token->u.rem); break;
      case T_STRING: /*{{{*/
      {
        size_t l=token->u.string->length;
        char *data=token->u.string->character;

        String_appendPrintf(s,"\"");
        while (l--)
        {
          if (*data=='"') String_appendPrintf(s,"\"");
          String_appendPrintf(s,"%c",*data);
          ++data;
        }
        String_appendPrintf(s,"\"");
        break;
      }
      /*}}}*/
      default: String_appendChars(s,table[token->type].text);
    }
    ns=table[token->type].space;
  } while (token++->type!=T_EOL);
  if (indent) *indent=nextindent;
  if (spaceto && s->length>oldlength) memset(s->character+oldlength,' ',s->length-oldlength);
  return s;
}
/*}}}*/
void Token_init(int b_c) /*{{{*/
{
#define PROPERTY(t,assoc,unary_priority,binary_priority,is_unary,is_binary) \
  Token_property[t]=(assoc<<8)|(unary_priority<<5)|(binary_priority<<2)|(is_unary<<1)|is_binary

  backslash_colon=b_c;
  PROPERTY(T_POW,  1,0,7,0,1);
  PROPERTY(T_MULT, 0,0,5,0,1);
  PROPERTY(T_DIV,  0,0,5,0,1);
  PROPERTY(T_IDIV, 0,0,5,0,1);
  PROPERTY(T_MOD,  0,0,5,0,1);
  PROPERTY(T_PLUS, 0,6,4,1,1);
  PROPERTY(T_MINUS,0,6,4,1,1);
  PROPERTY(T_LT,   0,0,3,0,1);
  PROPERTY(T_LE,   0,0,3,0,1);
  PROPERTY(T_EQ,   0,0,3,0,1);
  PROPERTY(T_GE,   0,0,3,0,1);
  PROPERTY(T_GT,   0,0,3,0,1);
  PROPERTY(T_NE,   0,0,3,0,1);
  PROPERTY(T_NOT,  0,2,0,1,0);
  PROPERTY(T_AND,  0,0,1,0,1);
  PROPERTY(T_OR,   0,0,0,0,1);
  PROPERTY(T_XOR,  0,0,0,0,1);
  PROPERTY(T_EQV,  0,0,0,0,1);
  PROPERTY(T_IMP,  0,0,0,0,1);
}
/*}}}*/
